dyn.load(file.path("src", paste("poisson_params", .Platform$dynlib.ext, sep="")))

top.sort <- function(model) {
  graph <- model$graph
  graph <- as.integer(graph)
  graph.size <- as.integer(dim(model$graph)[1])
  parents.num <- as.integer(tabulate(model$edges[,"to"], length(model$seqs)))
  queue <- as.integer(c(model$max.seqs,
                        rep(0, length(model$seqs) - length(model$max.seqs))))
  .C("top_sort",
     graph, graph.size,
     parents.num,
     queue, as.integer(length(model$max.seqs)),
     NAOK=TRUE,
     DUP=FALSE)
  queue
}

find.poisson.params <- function(model, cutting.weights, in.weights) {
  if(!is.integer(model$graph)) stop("find.poisson.params: non-integer model$graph!")
  graph.size <- as.integer(dim(model$graph)[1])
  if(!is.integer(model$topo.order)) stop("find.poisson.params: non-integer model$topo.order!")
  if(!is.double(cutting.weights)) stop("find.poisson.params: non-double cutting.weights!")
  poisson.params <- rep(NaN, length(model$seqs))
  if(!is.double(poisson.params)) stop("find.poisson.params: non-double poisson.params!")
  flow <- rep(as.double(0), length(model$seqs))
  flow[model$max.seqs] <- in.weights
  if(!is.double(flow)) stop("find.poisson.params: non-double flow!")
  .C("find_poisson_params",
     model$graph, graph.size,
     model$topo.order,
     cutting.weights,
     poisson.params,
     flow,
     NAOK=TRUE,
     DUP=FALSE)
  list(flow=flow, poisson.params=poisson.params)
}

# build the cleavage graph based on the peptides
# represented by the vector of strings seqs
# min.len is the minimal length of the sequences used in the graph
# returns a list with components:
# $alphabet - the letters used in seqs
# $seqs - all substrings of strings in seqs
# $lens - lengths of $seqs
# $max.seqs - indices in $seqs of maximal substrings (i.e. precursor peptides)
# $graph - has dimnames=list(NULL, c("left","right"), c("seq","letter"))
# $graph[i,"left","seq"] = k, iff cutting the letter $alphabet[$graph[i,"left","letter"]] from the left from seqs[i] gives seqs[k]
# $graph[i,"left","seq"] = NA, iff seqs[i] has length min.len (but "letter" is properly set anyway)
# analogously for "right", but the letter being cut is $alphabet[$graph[i,"left","letter"] - length($alphabet)]
# $edges - has dimnames=list(NULL, c("from","to","letter"))
# same info as in graph, but presented differently
# $topo.order - $seqs in topological order
make.model <- function(seqs, min.len=2) {
  alphabet <- unique(strsplit(paste(seqs, collapse=""), "")[[1]])
  f <- function(seq) {
    dummy <- diag(nchar(seq, type="chars"))
    ut <- upper.tri(dummy, diag=TRUE)
    substring(seq, row(dummy)[ut], col(dummy)[ut])
  }
  subseqs <- unique(unlist(sapply(seqs, f, simplify=FALSE, USE.NAMES=FALSE)))
  lens <- nchar(subseqs, type="chars")
  subseqs <- subseqs[lens >= min.len]
  lens <- lens[lens >= min.len]
  graph <- array(c(match(c(substring(subseqs, 2),
                           substring(subseqs, 1, lens-1)),
                         subseqs),
                   match(c(substring(subseqs, 1, 1),
                           substring(subseqs, lens, lens)),
                         alphabet)),
                 dim=c(length(subseqs), 2, 2),
                 dimnames=list(NULL, c("left","right"), c("seq","letter")))
  graph[,"right","letter"] <- graph[,"right","letter"] + length(alphabet)
  max.seqs <- setdiff(1:length(subseqs), graph[,,"seq"])
  edges <- cbind(1:length(subseqs),
                 as.vector(graph[,,"seq"]),
                 as.vector(graph[,,"letter"]))
  edges <- edges[complete.cases(edges),, drop=FALSE]
  dimnames(edges) <- list(NULL, c("from", "to", "letter"))
  model <- list(alphabet=alphabet, seqs=subseqs, lens=lens, max.seqs=max.seqs, graph=graph, edges=edges)
  topo.order <- top.sort(model)
  model[["topo.order"]] <- topo.order
  model
}

# make a vector of readings in the nodes of the cleavage graph
# poisson.params are the lambdas from the paper
# sd is the tau from the paper
# correct.reading.pr is the q from the paper
make.data <- function(poisson.params, sd=0.2, correct.reading.pr=0.7) {
  k <- length(poisson.params)
  poisson.params[is.na(poisson.params)] <- 0
  data <- rpois(k, poisson.params)
  data[data==0] <- NA #zeby nie miec klopotu jak z poissona zostalo wylosowane 0
  nas <- is.na(data)
  data[!nas] <- rlnorm(sum(!nas), log(data[!nas]), sd)
  data[data==0] <- NA
  nas <- is.na(data)
  incorrect <- !rbinom(sum(!nas), 1, correct.reading.pr)
  data[which(!nas)[incorrect]] <-
    1 + rlnorm(sum(incorrect),
               log(1 + rpois(sum(incorrect),
                             sample(poisson.params[poisson.params!=0],
                                    sum(incorrect), replace=TRUE))),
               sd)
  list(data=data, incorrect=incorrect)
}

par.log.f.exp <- function(x, y, FUN) {
  FUN <- match.fun(FUN)
  m <- pmax(x, y)
  log(FUN(exp(x-m), exp(y-m))) + m
}

rdirichlet <- function(a) {
  v <- rgamma(length(a), a)
  v / sum(v)
}

ddirichlet <- function(x, alpha, log=FALSE) {
  d <- lgamma(sum(alpha)) - sum(lgamma(alpha)) + sum((alpha - 1) * log(x))
  ifelse(log, d, exp(d))
}

# (y->x)/(x->y)
trans.ratio <- function(x, y, scale, log=FALSE) {
  d <- - sum(lgamma(scale * y + 1)) + sum(lgamma(scale * x + 1)) + scale * sum(log((x^y)/(y^x)))
  ifelse(log, d, exp(d))
}

estimate.mh <- function(x, model, steps,
                        step.kind.pr=c(1, 1, 1),
                        c.step.scale=8e1, # the c parameter from the paper for changing b
                        i.step.scale=8e1, # the c parameter from the paper for changing b_\star
                        e.step.sd=0.05, # the d parameter from the paper
                        cutting.weights.prior.params=rep(2, 2*length(model$alphabet)), # the B parameter from the paper
                        in.weights.prior.params=rep(2, length(model$max.seqs)), # the B_\star parameter from the paper
                        find.pp=find.poisson.params,
                        correct.reading.prior, # the q parameter from the paper
                        background, # the bg density from the paper, but in log-space (i.e. the density of the logarithm of the peak heights)
                        shape.prior=0, # the S_shape parameter from the paper
                        rate.prior=0, # the S_rate parameter from the paper
                        error.sd.prior=0.2, # the tau parameter from the paper
                        record.from=1, # will start recording the results from step number record.from
                        verbose=FALSE, # print step no and score for accepted transitions
                        input.permutation,
                        prev.result=NULL) {
  if(!missing(input.permutation))
    x <- x[input.permutation]
  observed <- which(!is.na(x))
  log.error.prior <- log(background(log(x[observed])))-log(x[observed])
  log.posterior <- function(y, poisson.params, cutting.weights, in.weights) {
    s <- sum(y) + shape.prior
    sum(par.log.f.exp(dlnorm(x[observed], log(y), error.sd.prior, log=TRUE) + log(correct.reading.prior),
                      log.error.prior + log(1-correct.reading.prior),
                      "+")) +
                        sum(y * log(poisson.params[observed])) -
                          sum(lgamma(y + 1)) +
                            lgamma(s) - s * log(sum(poisson.params[observed]) + rate.prior) +
                              ddirichlet(cutting.weights, cutting.weights.prior.params, log=TRUE) +
                                ddirichlet(in.weights, in.weights.prior.params, log=TRUE)
  }
  cutting.weights <- rdirichlet(cutting.weights.prior.params)
  in.weights <- rdirichlet(in.weights.prior.params)
  y <- rlnorm(length(observed), log(x[observed]), error.sd.prior) # co gdy istniej i: x[i]==0? co gdy istnieje i: y[i]==0?
  if(!is.null(prev.result)) {
    cutting.weights <- prev.result$cutting.weights[nrow(prev.result$cutting.weights),]
    in.weights <- prev.result$in.weights[nrow(prev.result$in.weights),]
    y <- prev.result$y[nrow(prev.result$y),]
  }
  poisson.params <- find.pp(model, cutting.weights, in.weights)$poisson.params
#  if(any(is.na(poisson.params) & apply(!is.na(model$graph[,,"seq"]), 1, any))) # terribly slow
#    stop("poisson.params is NaN!")
  score <- log.posterior(y, poisson.params, cutting.weights, in.weights)
  if(is.na(score)) stop("score is NaN!")
  acc.times <- 0
  acc.step.kind <- 0
  acc.scores <- score
  acc.cutting.weights <- cutting.weights
  acc.in.weights <- in.weights
  acc.y <- y
  for(t in 1:steps) {
    kind <- sample(3, 1, prob=step.kind.pr)
    cutting.weights.prop <- cutting.weights
    in.weights.prop <- in.weights
    poisson.params.prop <- poisson.params
    y.prop <- y
    log.trans <- 0
    if(kind == 1 && length(cutting.weights) > 1) {
      v <- sample(length(cutting.weights), 2)
      s <- sum(cutting.weights[v])
      repeat {
        cutting.weights.prop[v] <- rdirichlet(c.step.scale * cutting.weights[v] / s + 1) * s
        if(all(cutting.weights.prop[v]!=0)) break
        if(verbose) print("retrying cutting.weights")
      }
      if(any(is.na(cutting.weights.prop))) stop("cutting.weights.prop is NaN!")
      #if(any(cutting.weights.prop==0)) stop("cutting.weights.prop is 0!")
      log.trans <- trans.ratio(cutting.weights[v]/s, cutting.weights.prop[v]/s, c.step.scale, log=TRUE)
    }
    else if(kind == 2 && length(in.weights) > 1) {
      v <- sample(length(in.weights), 2)
      s <- sum(in.weights[v])
      repeat {
        in.weights.prop[v] <- rdirichlet(i.step.scale * in.weights[v] / s + 1) * s
        if(all(in.weights.prop[v]!=0)) break
        if(verbose) print("retrying in.weights")
      }
      if(any(is.na(in.weights.prop))) stop("in.weights.prop is NaN!")
      #if(any(in.weights.prop==0)) stop("in.weights.prop is 0!")
      log.trans <- trans.ratio(in.weights[v]/s, in.weights.prop[v]/s, i.step.scale, log=TRUE)
    }
    else if(kind == 3) {
      i <- sample(length(observed), 1)
      repeat {
        y.prop[i] <- rlnorm(1, log(y[i]), e.step.sd)
        if(y.prop[i]!=0) break
        if(verbose) print("retrying y")
      }
    }
    if(is.na(log.trans)) stop("log.trans is NaN!")
    if(!(kind %in% 3)) poisson.params.prop <- find.pp(model, cutting.weights.prop, in.weights.prop)$poisson.params
#    if(any(is.na(poisson.params.prop) & apply(!is.na(model$graph[,,"seq"]), 1, any))) # terribly slow
#      stop("poisson.params.prop is NaN!")
    score.prop <- log.posterior(y.prop, poisson.params.prop, cutting.weights.prop, in.weights.prop)
    if(is.na(score.prop)) stop("score.prop is NaN!")
    if(log(runif(1)) < log(8, base=8) * (score.prop - score + log.trans)) {
      cutting.weights <- cutting.weights.prop
      in.weights <- in.weights.prop
      poisson.params <- poisson.params.prop
      y <- y.prop
      score <- score.prop
      if(t >= record.from) {
        acc.times <- c(acc.times, t)
        acc.step.kind <- c(acc.step.kind, kind)
        acc.scores <- c(acc.scores, score)
        acc.cutting.weights <- c(acc.cutting.weights, cutting.weights)
        acc.in.weights <- c(acc.in.weights, in.weights)
        acc.y <- c(acc.y, y)
      }
      if(verbose) print(c(t, score))
    }
  }
  params <- list(steps=steps,
                 step.kind.pr=step.kind.pr,
                 c.step.scale=c.step.scale,
                 i.step.scale=i.step.scale,
                 e.step.sd=e.step.sd,
                 cutting.weights.prior.params=cutting.weights.prior.params,
                 in.weights.prior.params=in.weights.prior.params,
                 correct.reading.prior=correct.reading.prior,
                 shape.prior=shape.prior,
                 rate.prior=rate.prior,
                 error.sd.prior=error.sd.prior)
  list(times=acc.times, step.kind=acc.step.kind, scores=acc.scores,
       cutting.weights=matrix(acc.cutting.weights, ncol=length(cutting.weights), nrow=length(acc.times), byrow=TRUE),
       in.weights=matrix(acc.in.weights, ncol=length(in.weights), nrow=length(acc.times), byrow=TRUE),
       y=matrix(acc.y, ncol=length(y), nrow=length(acc.times), byrow=TRUE),
       params=params)
}
